From d9d58fc7d006f8fdb37d5ab9119062c5e0083c14 Mon Sep 17 00:00:00 2001
From: Bharath Rupireddy <bharath.rupireddyforpostgres@gmail.com>
Date: Tue, 8 Feb 2022 05:19:37 +0000
Subject: [PATCH v1] Refactor CheckpointWriteDelay()

---
 src/backend/postmaster/checkpointer.c | 36 +++++++--------------------
 src/backend/storage/buffer/bufmgr.c   |  6 +++--
 2 files changed, 13 insertions(+), 29 deletions(-)

diff --git a/src/backend/postmaster/checkpointer.c b/src/backend/postmaster/checkpointer.c
index 23f691cd47..e1f0ccd561 100644
--- a/src/backend/postmaster/checkpointer.c
+++ b/src/backend/postmaster/checkpointer.c
@@ -162,7 +162,6 @@ static pg_time_t last_xlog_switch_time;
 static void HandleCheckpointerInterrupts(void);
 static void CheckArchiveTimeout(void);
 static bool IsCheckpointOnSchedule(double progress);
-static bool ImmediateCheckpointRequested(void);
 static bool CompactCheckpointerRequestQueue(void);
 static void UpdateSharedMemoryConfig(void);
 
@@ -651,25 +650,6 @@ CheckArchiveTimeout(void)
 	}
 }
 
-/*
- * Returns true if an immediate checkpoint request is pending.  (Note that
- * this does not check the *current* checkpoint's IMMEDIATE flag, but whether
- * there is one pending behind it.)
- */
-static bool
-ImmediateCheckpointRequested(void)
-{
-	volatile CheckpointerShmemStruct *cps = CheckpointerShmem;
-
-	/*
-	 * We don't need to acquire the ckpt_lck in this case because we're only
-	 * looking at a single flag bit.
-	 */
-	if (cps->ckpt_flags & CHECKPOINT_IMMEDIATE)
-		return true;
-	return false;
-}
-
 /*
  * CheckpointWriteDelay -- control rate of checkpoint
  *
@@ -688,17 +668,19 @@ CheckpointWriteDelay(int flags, double progress)
 {
 	static int	absorb_counter = WRITES_PER_ABSORB;
 
-	/* Do nothing if checkpoint is being executed by non-checkpointer process */
-	if (!AmCheckpointerProcess())
-		return;
-
 	/*
-	 * Perform the usual duties and take a nap, unless we're behind schedule,
-	 * in which case we just try to catch up as quickly as possible.
+	 * Perform the usual duties and take a nap, unless we're behind schedule or
+	 * an immediate checkpoint request is pending, in which case we just try to
+	 * catch up as quickly as possible.
+	 *
+	 * We don't need to acquire the ckpt_lck while reading ckpt_flags from
+	 * checkpointer shared memory, because we're only looking at a single flag
+	 * bit.
 	 */
 	if (!(flags & CHECKPOINT_IMMEDIATE) &&
 		!ShutdownRequestPending &&
-		!ImmediateCheckpointRequested() &&
+		!(((volatile CheckpointerShmemStruct *) CheckpointerShmem)->ckpt_flags &
+		  CHECKPOINT_IMMEDIATE) &&
 		IsCheckpointOnSchedule(progress))
 	{
 		if (ConfigReloadPending)
diff --git a/src/backend/storage/buffer/bufmgr.c b/src/backend/storage/buffer/bufmgr.c
index f5459c68f8..859aa0c867 100644
--- a/src/backend/storage/buffer/bufmgr.c
+++ b/src/backend/storage/buffer/bufmgr.c
@@ -2172,11 +2172,13 @@ BufferSync(int flags)
 		}
 
 		/*
-		 * Sleep to throttle our I/O rate.
+		 * Sleep to throttle our I/O rate. Do nothing if checkpoint is being
+		 * executed by non-checkpointer process.
 		 *
 		 * (This will check for barrier events even if it doesn't sleep.)
 		 */
-		CheckpointWriteDelay(flags, (double) num_processed / num_to_scan);
+		if (AmCheckpointerProcess())
+			CheckpointWriteDelay(flags, (double) num_processed / num_to_scan);
 	}
 
 	/* issue all pending flushes */
-- 
2.25.1

